@@ -3,6 +3,8 @@
 # SPDX-License-Identifier: GPL-3.0-or-later
 # Copyright (C) 2018-2021 Purism SPC
 
+VERSION = "0.0.5"
+
 import argparse
 import datetime
 import hashlib
@@ -22,9 +24,9 @@ import yaml
 
 try:
     import jenkins
+    have_jenkins = True
 except ImportError:
-    logging.error("Cannot load Jenkins\nTry running: sudo apt install python3-jenkins")
-    exit(1)
+    have_jenkins = False
 
 try:
     import coloredlogs
@@ -38,16 +40,17 @@ from urllib.parse import urljoin
 IMAGES = {
     'stable': {
         'url': 'https://storage.puri.sm/librem5/images/',
+        'u-boot': 'https://storage.puri.sm/librem5/u-boot/',
     }
 }
 
-JENKINS = 'https://arm01.puri.sm'
+JENKINS = 'https://build.puri.sm'
 BOARD_TYPE = 'librem5r4'
 VALID_PHONE_BOARD_TYPES = ['librem5r2', 'librem5r3', 'librem5r4']
 VALID_DEVKIT_BOARD_TYPES = ['devkit']
-DIST = 'byzantium'
+DIST = 'current'
 IMAGE = '{}.img'
-IMAGE_VARIANT = 'plain'
+IMAGE_VARIANT = 'luks'
 META_YAML = 'files/meta.yml'
 IMAGE_JOB_NAME = 'Images/Image Build'
 UBOOT = 'u-boot-{}.imx'
@@ -67,10 +70,16 @@ SDPV: jump
 # This command will be run when ROM support stream mode
 SDPS: boot -f {uboot}
 SDPU: delay 1000
-#FB: ucmd setenv fastboot_buffer 0x43000000
+FB: ucmd setenv fastboot_buffer 0x43000000
+FB: download -f {uboot}
+FB: ucmd mmc dev 0 1
+FB: ucmd mmc write 0x43000000 0x42 0x1000
+FB: ucmd mmc partconf 0 0 {bootpart} 0
+FB: ucmd mmc dev 0 0
 FB: ucmd setenv fastboot_dev mmc
 FB: ucmd setenv mmcdev 0
 FB: flash -raw2sparse all {image}
+FB: reboot
 FB: Done
 '''
 BLOCK_SIZE = 8192
@@ -86,6 +95,15 @@ ATTR{idVendor}=="316d", ATTR{idProduct}=="4c05", GROUP+="plugdev", TAG+="uaccess
 LABEL="librem5_devkit_rules_end"
 '''
 
+# python3-urllib 2.x breaks python3-jenkins <1.8.1; it no longer accepts
+# socket._GLOBAL_DEFAULT_TIMEOUT as the default timeout.  This combination of
+# packages appears in Ubuntu 24.04 LTS and Debian Trixie (as of 2024-12-11).
+# Work around it by specifying our own timeout value.
+#
+# python-jenkins commit: https://opendev.org/jjb/python-jenkins/commit/ba9f06e0c2381f6506a92e9c0553b2f475975a18
+# python-jenkins bug: https://bugs.launchpad.net/python-jenkins/+bug/2018567
+JENKINS_TIMEOUT = 60
+
 
 class VerifyImageException(Exception):
     pass
@@ -217,7 +235,7 @@ def download_image(url, target, attempts):
 
 
 def find_image_jenkins(jobname, type, variant, dist):
-    server = jenkins.Jenkins(JENKINS)
+    server = jenkins.Jenkins(JENKINS, timeout=JENKINS_TIMEOUT)
     logging.info("Looking for {} {} {} image".format(type, variant, dist))
     try:
         info = server.get_job_info(jobname)
@@ -274,8 +292,35 @@ def find_image(jobname, board, variant, dist, stable):
         return find_image_jenkins(jobname, board, variant, dist)
 
 
-def find_uboot(jobname):
-    server = jenkins.Jenkins(JENKINS)
+def find_uboot_stable(board):
+    remote = IMAGES['stable']
+    logging.info("Looking for {} u-boot".format(board))
+    found = None
+
+    path = "latest/"
+    url = urljoin(remote['u-boot'], f"{path}/artifact/output/uboot-{board}/{UBOOT.format(board)}")
+    try:
+        resp = requests.head(url, timeout=10)
+        if resp.ok:
+            d = datetime.datetime.strptime(resp.headers['Last-Modified'],
+                                           '%a, %d %b %Y %H:%M:%S %Z')
+            return {
+                'url': urljoin(remote['u-boot'], path),
+                'timestamp': d.timestamp() * 1000,
+                'id': '"stable"',
+                'description': f"Last stable {board} build",
+            }
+        resp.raise_for_status()
+
+    except (requests.exceptions.ConnectionError,
+            requests.exceptions.Timeout):
+        logging.error("Failed to find u-boot at {} - connection failed".format(url))
+        found = None
+    return found
+
+
+def find_uboot_jenkins(jobname):
+    server = jenkins.Jenkins(JENKINS, timeout=JENKINS_TIMEOUT)
 
     try:
         info = server.get_job_info(jobname)
@@ -294,6 +339,13 @@ def find_uboot(jobname):
     return found
 
 
+def find_uboot(jobname, board, stable):
+    if stable:
+        return find_uboot_stable(board)
+    else:
+        return find_uboot_jenkins(jobname)
+
+
 def download_uboot(url, target):
     logging.info("Downloading uboot from {}".format(url))
     resp = requests.get(url, stream=True)
@@ -307,10 +359,11 @@ def download_uboot(url, target):
             bar.update(n=len(data))
 
 
-def write_uuu_script(target, image, uboot, mods):
+def write_uuu_script(target, image, uboot, bootpart, mods):
     with open(target, 'w+') as f:
         f.write(UUU_SCRIPT_TMPL.format(image=os.path.basename(image),
                                        uboot=os.path.basename(uboot),
+                                       bootpart=bootpart,
                                        mods=mods))
 
 
@@ -412,39 +465,53 @@ def setup_udev():
 def main():
     uuu_mods = ''
     parser = argparse.ArgumentParser(description='Flash a Librem 5 or Librem 5 Devkit.')
+    parser.add_argument('--version', action='version', version=VERSION)
     parser.add_argument('--dir', type=str, default=None,
-                        help='Download files to dir (instead of a temporary directory)')
-    parser.add_argument('--dist', type=str, default=DIST,
-                        help="Download an image for this distribution, default is '{}'".format(DIST))
+                        help='download files to DIR (instead of a temporary directory)')
     parser.add_argument('--skip-cleanup', action='store_true', default=False,
-                        help='Skip temporary directory cleanup')
+                        help='skip temporary directory cleanup')
     parser.add_argument('--skip-flash', action='store_true', default=False,
-                        help='Do all the preparations but don\'t flash')
-    parser.add_argument('--download-attempts', type=int, default=RETRIES,
-                        help="Maximum number of attempts to resume "
-                        "image download. 0: unlimited, default is {}".format(RETRIES))
+                        help='do all the preparations but don\'t flash')
     parser.add_argument('--skip-download', action='store_true', default=False,
-                        help='Use a local image and uboot instead of downloading')
-    parser.add_argument('--variant', choices=['plain', 'luks'], default=IMAGE_VARIANT,
-                        help="Image variant to download ( plain, luks ), "
-                        "default is '{}'".format(IMAGE_VARIANT))
-    parser.add_argument('--board', choices=VALID_DEVKIT_BOARD_TYPES + VALID_PHONE_BOARD_TYPES, default=None,
-                        help="Type of the board to download ( devkit, librem5r2, librem5r3, librem5r4 ) "
-                        "default is 'auto'")
-    parser.add_argument('--stable', action='store_true', default=False,
-                        help="Whether to grab the latest stable image, otherwise the dev snapshot is fetched")
+                        help='use already downloaded files (specified with --dir) instead of downloading')
+    parser.add_argument('--download-attempts', type=int, default=RETRIES,
+                        help="maximum number of attempts to resume "
+                        "image download. 0: unlimited, default is {}".format(RETRIES))
 
-    group = parser.add_argument_group(title='Testing and debugging options')
+    group = parser.add_argument_group(title='image options')
+    group.add_argument('--dist', type=str, default=DIST,
+                       help="download an image for this distribution ( byzantium, crimson, dawn, ... ) "
+                       "instead of the current release")
+    group.add_argument('--variant', choices=['plain', 'luks'], default=IMAGE_VARIANT,
+                       help="image variant to download ( plain, luks ), "
+                       "default is '{}'".format(IMAGE_VARIANT))
+    group.add_argument('--board', choices=VALID_DEVKIT_BOARD_TYPES + VALID_PHONE_BOARD_TYPES, default=None,
+                       help="type of the board to download ( devkit, librem5r2, librem5r3, librem5r4 ) "
+                       "default is 'auto'")
+    group.add_argument('--stable', action='store_true', default=True,
+                       help="download the latest stable image (default, use --snapshot to override)")
+    group.add_argument('--snapshot', action='store_true', default=False,
+                       help="download the latest CI snapshot instead of \"blessed\" images "
+                       "(takes precedence over --stable)")
+    group.add_argument('--image', type=str,
+                       help="flash the given uncompressed image file instead of downloading one")
+    group.add_argument('--uboot', type=str,
+                       help="use and flash the given bootloader file instead of downloading one")
+
+    group = parser.add_argument_group(title='testing and debugging options')
     group.add_argument('--debug', action="store_true", default=False,
-                       help='Enable debug output')
+                       help='enable debug output')
     group.add_argument('--image-job', type=str, default=IMAGE_JOB_NAME,
                        help='Jenkins job to download the image from')
     group.add_argument('--uboot-job', type=str,
                        help='Jenkins job to download the uboot from')
+    group.add_argument('--embedded-boot', action='store_true', default=False,
+                       help="set the bootloader embedded in the flashed image as bootable "
+                       "instead of using separate boot partition")
 
-    group = parser.add_argument_group(title='Initial setup options')
+    group = parser.add_argument_group(title='initial setup options')
     group.add_argument('--udev', action="store_true", default=False,
-                       help='Setup udev rules to flash without root permissions')
+                       help='setup udev rules to flash without root permissions')
     args = parser.parse_args()
 
     level = logging.DEBUG if args.debug else logging.INFO
@@ -456,38 +523,49 @@ def main():
     if args.udev:
         return setup_udev()
 
+    if args.snapshot:
+        args.stable = False
+        if not have_jenkins:
+            logging.error("Cannot load Jenkins\nTry running: sudo apt install python3-jenkins")
+            return 1
+
     board = args.board if args.board else get_board(sys.argv[0])
 
     # Check available downloads upfront so it's less likely we fail
     # later:
-    image_ref = find_image(args.image_job, board, args.variant, args.dist, args.stable)
-    if image_ref:
-        image_ref['ts'] = datetime.datetime.fromtimestamp(image_ref['timestamp'] / 1000).strftime('%c')
-        logging.info("Found disk image Build {id} '{description}' from {ts}".format(**image_ref))
-    else:
-        logging.error("No matching image found")
-        return 1
+    if not args.image:
+        image_ref = find_image(args.image_job, board, args.variant, args.dist, args.stable)
+        if image_ref:
+            image_ref['ts'] = datetime.datetime.fromtimestamp(image_ref['timestamp'] / 1000).strftime('%c')
+            logging.info("Found disk image Build {id} '{description}' from {ts}".format(**image_ref))
+        else:
+            logging.error("No matching image found")
+            return 1
 
-    if args.uboot_job:
-        uboot_ref = find_uboot(args.uboot_job)
-        uboot_board = board
-    else:
-        # uboot builds don't carry board revisions (yet?)
-        uboot_board = board[:-2] if re.match('librem5r[0-9]$', board) else board
-        uboot_ref = find_uboot(UBOOT_JOB_NAME.format(uboot_board))
+    uboot_board = board
 
-    if uboot_ref:
-        uboot_ref['ts'] = datetime.datetime.fromtimestamp(uboot_ref['timestamp'] / 1000).strftime('%c')
-        logging.info("Found uboot Build {id} from {ts}".format(**uboot_ref))
-    else:
-        logging.error("No matching uboot found")
-        return 1
+    if not args.uboot:
+        if args.snapshot and args.uboot_job:
+            uboot_ref = find_uboot_jenkins(args.uboot_job)
+            uboot_board = board
+        else:
+            # uboot builds don't carry board revisions (yet?)
+            uboot_board = board[:-2] if re.match('librem5r[0-9]$', board) else board
+            uboot_ref = find_uboot(UBOOT_JOB_NAME.format("librem5_mainline" if uboot_board == "librem5" else uboot_board), uboot_board, args.stable)
+
+        if uboot_ref:
+            uboot_ref['ts'] = datetime.datetime.fromtimestamp(uboot_ref['timestamp'] / 1000).strftime('%c')
+            logging.info("Found uboot Build {id} from {ts}".format(**uboot_ref))
+        else:
+            logging.error("No matching uboot found")
+            return 1
 
     if board in ["librem5r3", 'librem5r4']:
         uuu_mods = "CFG: SDP: -chip MX8MQ -compatible MX8MQ -vid 0x316d -pid 0x4c05"
 
     dirprefix = 'tmp_' + os.path.basename(sys.argv[0]) + '_'
     outdir = args.dir if args.dir is not None else tempfile.mkdtemp(prefix=dirprefix, dir='.')
+
     try:
         logging.info("Downloading to {}".format(outdir))
         if args.dir == outdir:
@@ -506,19 +584,32 @@ def main():
                 return 1
             logging.info(f"Skipping download and using local image '{image_target}'")
         else:
-            download_image(urljoin(image_ref['url'], 'artifact/{}.xz').format(IMAGE.format(board)),
-                           image_target, args.download_attempts)
-            download_uboot(urljoin(uboot_ref['url'],
-                           'artifact/output/uboot-{}/{}'.format(uboot_board, UBOOT.format(uboot_board))),
-                           uboot_target)
-        write_uuu_script(uuu_target, image_target, uboot_target, uuu_mods)
+            if args.image:
+                if not os.path.exists(args.image):
+                    logging.error(f"Given image {args.image} does not exist")
+                    return 1
+                os.symlink(os.path.abspath(args.image), image_target)
+            else:
+                download_image(urljoin(image_ref['url'], 'artifact/{}.xz').format(IMAGE.format(board)),
+                               image_target, args.download_attempts)
+
+            if args.uboot:
+                if not os.path.exists(args.uboot):
+                    logging.error(f"Given bootloader {args.uboot} does not exist")
+                    return 1
+                os.symlink(os.path.abspath(args.uboot), uboot_target)
+            else:
+                download_uboot(urljoin(uboot_ref['url'],
+                               'artifact/output/uboot-{}/{}'.format(uboot_board, UBOOT.format(uboot_board))),
+                               uboot_target)
+        write_uuu_script(uuu_target, image_target, uboot_target, 0 if args.embedded_boot else 1, uuu_mods)
         if not args.skip_flash:
             if board == "devkit":
                 wait_for_librem(librem_msg="Please connect your devkit now.\nSearching...")
             else:
                 wait_for_librem()
             flash_image(uuu_target, args.debug)
-            print("Flashing complete. You can now turn the device off.")
+            print("Flashing complete.")
     except VerifyImageException as e:
         logging.error(e)
         return 1
